/*
  This file is part of MADNESS.

  Copyright (C) 2007,2010 Oak Ridge National Laboratory

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

  For more information please contact:

  Robert J. Harrison
  Oak Ridge National Laboratory
  One Bethel Valley Road
  P.O. Box 2008, MS-6367

  email: harrisonrj@ornl.gov
  tel:   865-241-3937
  fax:   865-572-0680


  $Id$
*/


#define TENSOR_CC



#ifdef STATIC
#  undef STATIC
#endif

#if HAVE_UNQUALIFIED_STATIC_DECL
#  define STATIC static
#else
// Cray X1 compiler won't instantiate static function templates (mxm*)
#  define STATIC
#endif

#include <madness/madness_config.h>
#include <madness/misc/ran.h>
#include <madness/tensor/tensor.h>
#include <madness/tensor/mxm.h>
#include <madness/tensor/aligned.h>

#include <stdlib.h>
#include <algorithm>
#include <complex>
#include <cmath>
#include <cstring>
#include <iostream>


/// \file tensor.cc
/// \brief Completes the implementation of Tensor and instantiates all specializations for fast compiles.
/// \ingroup tensor
/// @{

namespace madness {

    template <class T> class Tensor;

    template <class T> class SliceTensor;

    template <class T> class GenTensor;

    template <class T> class SliceGenTensor;

    std::ostream& operator<<(std::ostream& stream, const Slice& s) {
        stream << "Slice(" << s.start << "," << s.end << "," << s.step << ")";
        return stream;
    }

    template<> float_complex Tensor<float_complex>::min(long* ind) const {
        TENSOR_EXCEPTION("cannot perform min on complex types",0,this);
        return 0;
    }
    template<> double_complex Tensor<double_complex>::min(long* ind) const {
        TENSOR_EXCEPTION("cannot perform min on complex types",0,this);
        return 0;
    }
    template<> float_complex Tensor<float_complex>::max(long* ind) const {
        TENSOR_EXCEPTION("cannot perform max on complex types",0,this);
        return 0;
    }
    template<> double_complex Tensor<double_complex>::max(long* ind) const {
        TENSOR_EXCEPTION("cannot perform max on complex types",0,this);
        return 0;
    }

    //#include "tensor_spec.h"
}
/// @}
