/*
   Copyright (c) 2009-2014, Jack Poulson
   All rights reserved.

   This file is part of Elemental and is under the BSD 2-Clause License, 
   which can be found in the LICENSE file in the root directory, or at 
   http://opensource.org/licenses/BSD-2-Clause
*/
#pragma once
#ifndef ELEM_HERMITIANEIG_HPP
#define ELEM_HERMITIANEIG_HPP

#include ELEM_HERMITIANTRIDIAG_INC
#include ELEM_HERMITIANTRIDIAGEIG_INC

// NOTE: HermitianEig declarations are in lapack-like/decomp/decl.hpp
#include "./HermitianEig/SDC.hpp"

#endif // ifndef ELEM_HERMITIANEIG_HPP
