template class TensorIterator<double,double,double>;
template class TensorIterator<double,double,double_complex>;
template class TensorIterator<double,double,float>;
template class TensorIterator<double,double,float_complex>;
template class TensorIterator<double,double,int>;
template class TensorIterator<double,double,long>;
template class TensorIterator<double,double_complex,double>;
template class TensorIterator<double,double_complex,double_complex>;
template class TensorIterator<double,double_complex,float>;
template class TensorIterator<double,double_complex,float_complex>;
template class TensorIterator<double,double_complex,int>;
template class TensorIterator<double,double_complex,long>;
template class TensorIterator<double,float,double>;
template class TensorIterator<double,float,double_complex>;
template class TensorIterator<double,float,float>;
template class TensorIterator<double,float,float_complex>;
template class TensorIterator<double,float,int>;
template class TensorIterator<double,float,long>;
template class TensorIterator<double,float_complex,double>;
template class TensorIterator<double,float_complex,double_complex>;
template class TensorIterator<double,float_complex,float>;
template class TensorIterator<double,float_complex,float_complex>;
template class TensorIterator<double,float_complex,int>;
template class TensorIterator<double,float_complex,long>;
template class TensorIterator<double,int,double>;
template class TensorIterator<double,int,double_complex>;
template class TensorIterator<double,int,float>;
template class TensorIterator<double,int,float_complex>;
template class TensorIterator<double,int,int>;
template class TensorIterator<double,int,long>;
template class TensorIterator<double,long,double>;
template class TensorIterator<double,long,double_complex>;
template class TensorIterator<double,long,float>;
template class TensorIterator<double,long,float_complex>;
template class TensorIterator<double,long,int>;
template class TensorIterator<double,long,long>;
template class TensorIterator<double_complex,double,double>;
template class TensorIterator<double_complex,double,double_complex>;
template class TensorIterator<double_complex,double,float>;
template class TensorIterator<double_complex,double,float_complex>;
template class TensorIterator<double_complex,double,int>;
template class TensorIterator<double_complex,double,long>;
template class TensorIterator<double_complex,double_complex,double>;
template class TensorIterator<double_complex,double_complex,double_complex>;
template class TensorIterator<double_complex,double_complex,float>;
template class TensorIterator<double_complex,double_complex,float_complex>;
template class TensorIterator<double_complex,double_complex,int>;
template class TensorIterator<double_complex,double_complex,long>;
template class TensorIterator<double_complex,float,double>;
template class TensorIterator<double_complex,float,double_complex>;
template class TensorIterator<double_complex,float,float>;
template class TensorIterator<double_complex,float,float_complex>;
template class TensorIterator<double_complex,float,int>;
template class TensorIterator<double_complex,float,long>;
template class TensorIterator<double_complex,float_complex,double>;
template class TensorIterator<double_complex,float_complex,double_complex>;
template class TensorIterator<double_complex,float_complex,float>;
template class TensorIterator<double_complex,float_complex,float_complex>;
template class TensorIterator<double_complex,float_complex,int>;
template class TensorIterator<double_complex,float_complex,long>;
template class TensorIterator<double_complex,int,double>;
template class TensorIterator<double_complex,int,double_complex>;
template class TensorIterator<double_complex,int,float>;
template class TensorIterator<double_complex,int,float_complex>;
template class TensorIterator<double_complex,int,int>;
template class TensorIterator<double_complex,int,long>;
template class TensorIterator<double_complex,long,double>;
template class TensorIterator<double_complex,long,double_complex>;
template class TensorIterator<double_complex,long,float>;
template class TensorIterator<double_complex,long,float_complex>;
template class TensorIterator<double_complex,long,int>;
template class TensorIterator<double_complex,long,long>;
template class TensorIterator<float,double,double>;
template class TensorIterator<float,double,double_complex>;
template class TensorIterator<float,double,float>;
template class TensorIterator<float,double,float_complex>;
template class TensorIterator<float,double,int>;
template class TensorIterator<float,double,long>;
template class TensorIterator<float,double_complex,double>;
template class TensorIterator<float,double_complex,double_complex>;
template class TensorIterator<float,double_complex,float>;
template class TensorIterator<float,double_complex,float_complex>;
template class TensorIterator<float,double_complex,int>;
template class TensorIterator<float,double_complex,long>;
template class TensorIterator<float,float,double>;
template class TensorIterator<float,float,double_complex>;
template class TensorIterator<float,float,float>;
template class TensorIterator<float,float,float_complex>;
template class TensorIterator<float,float,int>;
template class TensorIterator<float,float,long>;
template class TensorIterator<float,float_complex,double>;
template class TensorIterator<float,float_complex,double_complex>;
template class TensorIterator<float,float_complex,float>;
template class TensorIterator<float,float_complex,float_complex>;
template class TensorIterator<float,float_complex,int>;
template class TensorIterator<float,float_complex,long>;
template class TensorIterator<float,int,double>;
template class TensorIterator<float,int,double_complex>;
template class TensorIterator<float,int,float>;
template class TensorIterator<float,int,float_complex>;
template class TensorIterator<float,int,int>;
template class TensorIterator<float,int,long>;
template class TensorIterator<float,long,double>;
template class TensorIterator<float,long,double_complex>;
template class TensorIterator<float,long,float>;
template class TensorIterator<float,long,float_complex>;
template class TensorIterator<float,long,int>;
template class TensorIterator<float,long,long>;
template class TensorIterator<float_complex,double,double>;
template class TensorIterator<float_complex,double,double_complex>;
template class TensorIterator<float_complex,double,float>;
template class TensorIterator<float_complex,double,float_complex>;
template class TensorIterator<float_complex,double,int>;
template class TensorIterator<float_complex,double,long>;
template class TensorIterator<float_complex,double_complex,double>;
template class TensorIterator<float_complex,double_complex,double_complex>;
template class TensorIterator<float_complex,double_complex,float>;
template class TensorIterator<float_complex,double_complex,float_complex>;
template class TensorIterator<float_complex,double_complex,int>;
template class TensorIterator<float_complex,double_complex,long>;
template class TensorIterator<float_complex,float,double>;
template class TensorIterator<float_complex,float,double_complex>;
template class TensorIterator<float_complex,float,float>;
template class TensorIterator<float_complex,float,float_complex>;
template class TensorIterator<float_complex,float,int>;
template class TensorIterator<float_complex,float,long>;
template class TensorIterator<float_complex,float_complex,double>;
template class TensorIterator<float_complex,float_complex,double_complex>;
template class TensorIterator<float_complex,float_complex,float>;
template class TensorIterator<float_complex,float_complex,float_complex>;
template class TensorIterator<float_complex,float_complex,int>;
template class TensorIterator<float_complex,float_complex,long>;
template class TensorIterator<float_complex,int,double>;
template class TensorIterator<float_complex,int,double_complex>;
template class TensorIterator<float_complex,int,float>;
template class TensorIterator<float_complex,int,float_complex>;
template class TensorIterator<float_complex,int,int>;
template class TensorIterator<float_complex,int,long>;
template class TensorIterator<float_complex,long,double>;
template class TensorIterator<float_complex,long,double_complex>;
template class TensorIterator<float_complex,long,float>;
template class TensorIterator<float_complex,long,float_complex>;
template class TensorIterator<float_complex,long,int>;
template class TensorIterator<float_complex,long,long>;
template class TensorIterator<int,double,double>;
template class TensorIterator<int,double,double_complex>;
template class TensorIterator<int,double,float>;
template class TensorIterator<int,double,float_complex>;
template class TensorIterator<int,double,int>;
template class TensorIterator<int,double,long>;
template class TensorIterator<int,double_complex,double>;
template class TensorIterator<int,double_complex,double_complex>;
template class TensorIterator<int,double_complex,float>;
template class TensorIterator<int,double_complex,float_complex>;
template class TensorIterator<int,double_complex,int>;
template class TensorIterator<int,double_complex,long>;
template class TensorIterator<int,float,double>;
template class TensorIterator<int,float,double_complex>;
template class TensorIterator<int,float,float>;
template class TensorIterator<int,float,float_complex>;
template class TensorIterator<int,float,int>;
template class TensorIterator<int,float,long>;
template class TensorIterator<int,float_complex,double>;
template class TensorIterator<int,float_complex,double_complex>;
template class TensorIterator<int,float_complex,float>;
template class TensorIterator<int,float_complex,float_complex>;
template class TensorIterator<int,float_complex,int>;
template class TensorIterator<int,float_complex,long>;
template class TensorIterator<int,int,double>;
template class TensorIterator<int,int,double_complex>;
template class TensorIterator<int,int,float>;
template class TensorIterator<int,int,float_complex>;
template class TensorIterator<int,int,int>;
template class TensorIterator<int,int,long>;
template class TensorIterator<int,long,double>;
template class TensorIterator<int,long,double_complex>;
template class TensorIterator<int,long,float>;
template class TensorIterator<int,long,float_complex>;
template class TensorIterator<int,long,int>;
template class TensorIterator<int,long,long>;
template class TensorIterator<long,double,double>;
template class TensorIterator<long,double,double_complex>;
template class TensorIterator<long,double,float>;
template class TensorIterator<long,double,float_complex>;
template class TensorIterator<long,double,int>;
template class TensorIterator<long,double,long>;
template class TensorIterator<long,double_complex,double>;
template class TensorIterator<long,double_complex,double_complex>;
template class TensorIterator<long,double_complex,float>;
template class TensorIterator<long,double_complex,float_complex>;
template class TensorIterator<long,double_complex,int>;
template class TensorIterator<long,double_complex,long>;
template class TensorIterator<long,float,double>;
template class TensorIterator<long,float,double_complex>;
template class TensorIterator<long,float,float>;
template class TensorIterator<long,float,float_complex>;
template class TensorIterator<long,float,int>;
template class TensorIterator<long,float,long>;
template class TensorIterator<long,float_complex,double>;
template class TensorIterator<long,float_complex,double_complex>;
template class TensorIterator<long,float_complex,float>;
template class TensorIterator<long,float_complex,float_complex>;
template class TensorIterator<long,float_complex,int>;
template class TensorIterator<long,float_complex,long>;
template class TensorIterator<long,int,double>;
template class TensorIterator<long,int,double_complex>;
template class TensorIterator<long,int,float>;
template class TensorIterator<long,int,float_complex>;
template class TensorIterator<long,int,int>;
template class TensorIterator<long,int,long>;
template class TensorIterator<long,long,double>;
template class TensorIterator<long,long,double_complex>;
template class TensorIterator<long,long,float>;
template class TensorIterator<long,long,float_complex>;
template class TensorIterator<long,long,int>;
template class TensorIterator<long,long,long>;
