MTXM_ENTRY(mTxm2)
.JLOOP2:
ZERO(C00)
.KLOOP2:
LOADBKJ
ABC(0,C00)
INCA
ABC2(0,C00)
INCA
sub     $2, NK
jnz     .KLOOP2
mov     C, B
STORE(C00)
INCC
NEXTJ
jnz     .JLOOP2
RETURN
MTXM_ENTRY(mTxm4)
.JLOOP4:
ZERO(C00)
ZERO(C01)
.KLOOP4:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
INCA
ABC2(0,C00)
ABC2(16,C01)
INCA
sub     $2, NK
jnz     .KLOOP4
mov     C, B
STORE(C00)
STORE(C01)
INCC
NEXTJ
jnz     .JLOOP4
RETURN
MTXM_ENTRY(mTxm6)
.JLOOP6:
ZERO(C00)
ZERO(C01)
ZERO(C02)
.KLOOP6:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
INCA
sub     $2, NK
jnz     .KLOOP6
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
INCC
NEXTJ
jnz     .JLOOP6
RETURN
MTXM_ENTRY(mTxm8)
.JLOOP8:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
.KLOOP8:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
INCA
sub     $2, NK
jnz     .KLOOP8
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
INCC
NEXTJ
jnz     .JLOOP8
RETURN
MTXM_ENTRY(mTxm10)
.JLOOP10:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
.KLOOP10:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
INCA
sub     $2, NK
jnz     .KLOOP10
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
STORE(C04)
INCC
NEXTJ
jnz     .JLOOP10
RETURN
#ifdef X86_64
MTXM_ENTRY(mTxm12)
.JLOOP12:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
.KLOOP12:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
INCA
sub     $2, NK
jnz     .KLOOP12
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
STORE(C04)
STORE(C05)
INCC
NEXTJ
jnz     .JLOOP12
RETURN
MTXM_ENTRY(mTxm14)
.JLOOP14:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
.KLOOP14:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
INCA
sub     $2, NK
jnz     .KLOOP14
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
STORE(C04)
STORE(C05)
STORE(C06)
INCC
NEXTJ
jnz     .JLOOP14
RETURN
MTXM_ENTRY(mTxm16)
.JLOOP16:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
.KLOOP16:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
INCA
sub     $2, NK
jnz     .KLOOP16
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
STORE(C04)
STORE(C05)
STORE(C06)
STORE(C07)
INCC
NEXTJ
jnz     .JLOOP16
RETURN
MTXM_ENTRY(mTxm18)
.JLOOP18:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
.KLOOP18:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
INCA
sub     $2, NK
jnz     .KLOOP18
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
STORE(C04)
STORE(C05)
STORE(C06)
STORE(C07)
STORE(C08)
INCC
NEXTJ
jnz     .JLOOP18
RETURN
MTXM_ENTRY(mTxm20)
.JLOOP20:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
ZERO(C09)
.KLOOP20:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
ABC(144,C09)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
ABC2(144,C09)
INCA
sub     $2, NK
jnz     .KLOOP20
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
STORE(C04)
STORE(C05)
STORE(C06)
STORE(C07)
STORE(C08)
STORE(C09)
INCC
NEXTJ
jnz     .JLOOP20
RETURN
MTXM_ENTRY(mTxm22)
.JLOOP22:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
ZERO(C09)
ZERO(C10)
.KLOOP22:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
ABC(144,C09)
ABC(160,C10)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
ABC2(144,C09)
ABC2(160,C10)
INCA
sub     $2, NK
jnz     .KLOOP22
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
STORE(C04)
STORE(C05)
STORE(C06)
STORE(C07)
STORE(C08)
STORE(C09)
STORE(C10)
INCC
NEXTJ
jnz     .JLOOP22
RETURN
MTXM_ENTRY(mTxm24)
.JLOOP24:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
ZERO(C09)
ZERO(C10)
ZERO(C11)
.KLOOP24:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
ABC(144,C09)
ABC(160,C10)
ABC(176,C11)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
ABC2(144,C09)
ABC2(160,C10)
ABC2(176,C11)
INCA
sub     $2, NK
jnz     .KLOOP24
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
STORE(C04)
STORE(C05)
STORE(C06)
STORE(C07)
STORE(C08)
STORE(C09)
STORE(C10)
STORE(C11)
INCC
NEXTJ
jnz     .JLOOP24
RETURN
MTXM_ENTRY(mTxm26)
.JLOOP26:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
ZERO(C09)
ZERO(C10)
ZERO(C11)
ZERO(C12)
.KLOOP26:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
ABC(144,C09)
ABC(160,C10)
ABC(176,C11)
ABC(192,C12)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
ABC2(144,C09)
ABC2(160,C10)
ABC2(176,C11)
ABC2(192,C12)
INCA
sub     $2, NK
jnz     .KLOOP26
mov     C, B
STORE(C00)
STORE(C01)
STORE(C02)
STORE(C03)
STORE(C04)
STORE(C05)
STORE(C06)
STORE(C07)
STORE(C08)
STORE(C09)
STORE(C10)
STORE(C11)
STORE(C12)
INCC
NEXTJ
jnz     .JLOOP26
RETURN
#endif
MTXM_ENTRY(TmTxm2)
.TJLOOP2:
ZERO(C00)
.TKLOOP2:
LOADBKJ
ABC(0,C00)
INCA
ABC2(0,C00)
INCA
sub     $2, NK
jnz     .TKLOOP2
mov     C, B
STORET(C00)
INCCT
NEXTJ
jnz     .TJLOOP2
RETURN
MTXM_ENTRY(TmTxm4)
.TJLOOP4:
ZERO(C00)
ZERO(C01)
.TKLOOP4:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
INCA
ABC2(0,C00)
ABC2(16,C01)
INCA
sub     $2, NK
jnz     .TKLOOP4
mov     C, B
STORET(C00)
STORET(C01)
INCCT
NEXTJ
jnz     .TJLOOP4
RETURN
MTXM_ENTRY(TmTxm6)
.TJLOOP6:
ZERO(C00)
ZERO(C01)
ZERO(C02)
.TKLOOP6:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
INCA
sub     $2, NK
jnz     .TKLOOP6
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
INCCT
NEXTJ
jnz     .TJLOOP6
RETURN
MTXM_ENTRY(TmTxm8)
.TJLOOP8:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
.TKLOOP8:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
INCA
sub     $2, NK
jnz     .TKLOOP8
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
INCCT
NEXTJ
jnz     .TJLOOP8
RETURN
MTXM_ENTRY(TmTxm10)
.TJLOOP10:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
.TKLOOP10:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
INCA
sub     $2, NK
jnz     .TKLOOP10
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
STORET(C04)
INCCT
NEXTJ
jnz     .TJLOOP10
RETURN
#ifdef X86_64
MTXM_ENTRY(TmTxm12)
.TJLOOP12:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
.TKLOOP12:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
INCA
sub     $2, NK
jnz     .TKLOOP12
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
STORET(C04)
STORET(C05)
INCCT
NEXTJ
jnz     .TJLOOP12
RETURN
MTXM_ENTRY(TmTxm14)
.TJLOOP14:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
.TKLOOP14:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
INCA
sub     $2, NK
jnz     .TKLOOP14
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
STORET(C04)
STORET(C05)
STORET(C06)
INCCT
NEXTJ
jnz     .TJLOOP14
RETURN
MTXM_ENTRY(TmTxm16)
.TJLOOP16:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
.TKLOOP16:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
INCA
sub     $2, NK
jnz     .TKLOOP16
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
STORET(C04)
STORET(C05)
STORET(C06)
STORET(C07)
INCCT
NEXTJ
jnz     .TJLOOP16
RETURN
MTXM_ENTRY(TmTxm18)
.TJLOOP18:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
.TKLOOP18:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
INCA
sub     $2, NK
jnz     .TKLOOP18
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
STORET(C04)
STORET(C05)
STORET(C06)
STORET(C07)
STORET(C08)
INCCT
NEXTJ
jnz     .TJLOOP18
RETURN
MTXM_ENTRY(TmTxm20)
.TJLOOP20:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
ZERO(C09)
.TKLOOP20:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
ABC(144,C09)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
ABC2(144,C09)
INCA
sub     $2, NK
jnz     .TKLOOP20
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
STORET(C04)
STORET(C05)
STORET(C06)
STORET(C07)
STORET(C08)
STORET(C09)
INCCT
NEXTJ
jnz     .TJLOOP20
RETURN
MTXM_ENTRY(TmTxm22)
.TJLOOP22:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
ZERO(C09)
ZERO(C10)
.TKLOOP22:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
ABC(144,C09)
ABC(160,C10)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
ABC2(144,C09)
ABC2(160,C10)
INCA
sub     $2, NK
jnz     .TKLOOP22
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
STORET(C04)
STORET(C05)
STORET(C06)
STORET(C07)
STORET(C08)
STORET(C09)
STORET(C10)
INCCT
NEXTJ
jnz     .TJLOOP22
RETURN
MTXM_ENTRY(TmTxm24)
.TJLOOP24:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
ZERO(C09)
ZERO(C10)
ZERO(C11)
.TKLOOP24:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
ABC(144,C09)
ABC(160,C10)
ABC(176,C11)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
ABC2(144,C09)
ABC2(160,C10)
ABC2(176,C11)
INCA
sub     $2, NK
jnz     .TKLOOP24
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
STORET(C04)
STORET(C05)
STORET(C06)
STORET(C07)
STORET(C08)
STORET(C09)
STORET(C10)
STORET(C11)
INCCT
NEXTJ
jnz     .TJLOOP24
RETURN
MTXM_ENTRY(TmTxm26)
.TJLOOP26:
ZERO(C00)
ZERO(C01)
ZERO(C02)
ZERO(C03)
ZERO(C04)
ZERO(C05)
ZERO(C06)
ZERO(C07)
ZERO(C08)
ZERO(C09)
ZERO(C10)
ZERO(C11)
ZERO(C12)
.TKLOOP26:
LOADBKJ
ABC(0,C00)
ABC(16,C01)
ABC(32,C02)
ABC(48,C03)
ABC(64,C04)
ABC(80,C05)
ABC(96,C06)
ABC(112,C07)
ABC(128,C08)
ABC(144,C09)
ABC(160,C10)
ABC(176,C11)
ABC(192,C12)
INCA
ABC2(0,C00)
ABC2(16,C01)
ABC2(32,C02)
ABC2(48,C03)
ABC2(64,C04)
ABC2(80,C05)
ABC2(96,C06)
ABC2(112,C07)
ABC2(128,C08)
ABC2(144,C09)
ABC2(160,C10)
ABC2(176,C11)
ABC2(192,C12)
INCA
sub     $2, NK
jnz     .TKLOOP26
mov     C, B
STORET(C00)
STORET(C01)
STORET(C02)
STORET(C03)
STORET(C04)
STORET(C05)
STORET(C06)
STORET(C07)
STORET(C08)
STORET(C09)
STORET(C10)
STORET(C11)
STORET(C12)
INCCT
NEXTJ
jnz     .TJLOOP26
RETURN
#endif
