/*
   Copyright (c) 2009-2014, Jack Poulson
   All rights reserved.

   This file is part of Elemental and is under the BSD 2-Clause License, 
   which can be found in the LICENSE file in the root directory, or at 
   http://opensource.org/licenses/BSD-2-Clause
*/
#pragma once
#ifndef ELEMENTAL_HPP
#define ELEMENTAL_HPP

#include "elemental-lite.hpp"

#include "elemental/core.hpp"
#include "elemental/blas-like.hpp"
#include "elemental/lapack-like.hpp"
#include "elemental/convex.hpp"
#include "elemental/control.hpp"
#include "elemental/matrices.hpp"

#endif // ifndef ELEMENTAL_HPP
