/*
  This file is part of MADNESS.

  Copyright (C) 2007,2010 Oak Ridge National Laboratory

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

  For more information please contact:

  Robert J. Harrison
  Oak Ridge National Laboratory
  One Bethel Valley Road
  P.O. Box 2008, MS-6367

  email: harrisonrj@ornl.gov
  tel:   865-241-3937
  fax:   865-572-0680
*/

/**
 \file world_task_queue.cc
 \brief Implementation of \c TaskInterface and \c WorldTaskQueue functions.
 \ingroup taskq
*/

#include <madness/world/world_task_queue.h>

namespace madness {

    bool TaskInterface::debug = false;

    // This is what thread pool will invoke
    void TaskInterface::run(const TaskThreadEnv& env) {
        MADNESS_ASSERT(world);
        MADNESS_ASSERT(completion);
        World* w = const_cast<World*>(world);
        if (debug) std::cerr << w->rank() << ": Task " << (void*) this << " is now running" << std::endl;
        run(*w, env);
        if (debug) std::cerr << w->rank() << ": Task " << (void*) this << " has completed" << std::endl;
    }

    WorldTaskQueue::WorldTaskQueue(World& world)
            : world(world)
            , me(world.rank()) {
        nregistered = 0;
    }

}  // namespace madness
