/*
   Copyright (c) 2009-2014, Jack Poulson
   All rights reserved.

   This file is part of Elemental and is under the BSD 2-Clause License, 
   which can be found in the LICENSE file in the root directory, or at 
   http://opensource.org/licenses/BSD-2-Clause
*/
#pragma once
#ifndef ELEM_SOLVE_HPP
#define ELEM_SOLVE_HPP

#include "./solve/decl.hpp"
#include "./solve/impl.hpp"

#endif // ifndef ELEM_SOLVE_HPP
